/* Copyright 2005 Garrett Rooney.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_template.h"

#include <apr_strings.h>
#include <apr_getopt.h>

#include <stdlib.h>

static void
usage ()
{
  fprintf (stderr, "usage: tmpl [-D name=val] <template>\n");
  exit (1);
}

int
main (int argc, const char * const argv[])
{
  apr_bucket_alloc_t *alloc;
  apr_bucket_brigade *out;
  apr_hash_t *environment;
  apr_template_t *tmpl;
  apr_status_t apr_err;
  apr_pool_t *pool;
  apr_getopt_t *o;
  const char *arg;
  char opt;

  apr_initialize ();

  atexit (apr_terminate);

  apr_pool_create (&pool, NULL);

  environment = apr_hash_make (pool);

  apr_getopt_init (&o, pool, argc, argv);

  for (;;)
    {
      apr_err = apr_getopt (o, "D:", &opt, &arg);
      if (apr_err == APR_EOF)
        break;
      else if (apr_err != APR_SUCCESS)
        usage ();
      else
        {
          switch (opt)
            {
              case 'D':
                {
                  apr_template_variable_t *v;
                  char *name, *tmp;
                 
                  name = apr_strtok (apr_pstrdup (pool, arg), "=", &tmp);
                  if (! name)
                    usage ();

                  v = apr_pcalloc (pool, sizeof (*v));

                  v->type = APR_TEMPLATE_SCALAR;

                  v->contents.s = apr_strtok (NULL, "=", &tmp);
                  if (! v->contents.s)
                    usage ();

                  apr_hash_set (environment, name, APR_HASH_KEY_STRING, v);
                }
                break;

              default:
                break;
            }
        }
    }

  if ((o->argc - o->ind) != 1)
    usage ();

  alloc = apr_bucket_alloc_create (pool);

  out = apr_brigade_create (pool, alloc);

  apr_err = apr_template_parse_file (&tmpl, argv[o->ind], pool);
  if (apr_err)
    goto cleanup;

  apr_err = apr_template_execute (tmpl, environment, out, pool);
  if (apr_err)
    goto cleanup;

  {
    apr_bucket_brigade *tb = apr_brigade_create (pool, alloc);

    for (;;)
      {
        char buffer[1024] = { 0 };

        apr_size_t buffsize = sizeof (buffer);

        if (APR_BRIGADE_EMPTY (out))
          break;

        apr_err = apr_brigade_split_line (tb,
                                          out,
                                          APR_BLOCK_READ,
                                          sizeof (buffer));
        if (apr_err)
          break;

        apr_brigade_flatten (tb, buffer, &buffsize);

        buffer[buffsize] = 0;

        printf ("%s", buffer);

        apr_brigade_cleanup (tb);
      }

    apr_brigade_destroy (tb);
  }

cleanup:

  if (apr_err)
    {
      char buffer[1024] = { 0 };

      fprintf (stderr, "error: %s\n", apr_strerror (apr_err,
                                                    buffer,
                                                    sizeof (buffer)));
      return 1;
    }

  apr_brigade_destroy (out);

  apr_bucket_alloc_destroy (alloc);

  apr_pool_destroy (pool);

  return 0;
}
